program SUMMA
!       Creates a 2d grid and do a matrix - matrix multiplication 
!       between to matrices of the same dimensions using the MPI interface!

include 'mpif.h'

integer,parameter:: DIMM     = 100
integer,parameter:: m        = DIMM
integer,parameter:: n        = DIMM
integer,parameter:: true     = 1
integer,parameter:: false    = 0

logical,parameter:: REORDER  = .true.   ! Reorder nodes if nessesary
integer,parameter:: CART_DIM = 2        ! Dimension of cartesian map
logical,parameter:: NO_WRAP  = .false.  ! No wrap around corners of the map

integer,parameter:: UPDOWN   = 0
integer,parameter:: SIDEWAYS = 1
integer,parameter:: LEFT     = 2
integer,parameter:: RIGHT    = 3
integer,parameter:: DOWN     = 1
integer,parameter:: UP       = 0
integer,parameter:: X        = 1
integer,parameter:: Y        = 0

real*8   :: t1;
integer:: my_id, numprocs, i,j,k;
integer:: my_x, my_y, dim_x, dim_y,temp,my_m,my_n,ierr;
integer:: mdim(0:99),ndim(0:99),mpos(0:99),npos(0:99);
integer:: cartsize(0:1),mycoord(0:1),link(0:3);
logical:: period(0:1)

real,dimension(:,:),allocatable::A,B,C
real,dimension(:),allocatable::workA,workB

integer:: cart_comm,sendnr
integer:: status(MPI_STATUS_SIZE)

call MPI_INIT(ierr)
call MPI_COMM_SIZE(MPI_COMM_WORLD,numprocs,ierr)
call MPI_COMM_RANK(MPI_COMM_WORLD,my_id,ierr)

! Build a cartesian grid with optimal size 
cartsize=(/0,0/)
call MPI_DIMS_CREATE(numprocs,2,cartsize, ierr)
dim_x=cartsize(X)
dim_y=cartsize(Y)

if (my_id==0) then
   print *,"Dimension:",DIMM
   print *,"Number of proc:",numprocs
   print *,"DIM x:%d  DIM y:",dim_x,dim_y
end if
!print *,"Proc ",my_id," ready!"
  
period=(/ NO_WRAP, NO_WRAP/)
call MPI_CART_CREATE(MPI_COMM_WORLD,CART_DIM,cartsize, &
     period,REORDER,cart_comm,ierr)

call MPI_COMM_RANK(MPI_COMM_WORLD,my_id,ierr)
call MPI_CART_COORDS(cart_comm,my_id,CART_DIM,mycoord,ierr)
call MPI_CART_SHIFT(cart_comm,SIDEWAYS,1,link(LEFT),link(RIGHT),ierr)
call MPI_CART_SHIFT(cart_comm,UPDOWN  ,1,link(UP)  ,link(DOWN) ,ierr)

my_x=mycoord(X)
my_y=mycoord(Y)
  
  
! Distribute matrices 
  
!     Compute the dimensions of the partitions of A over the proc.
!     Distrubute the matrix in blocks
!     ndim(my_x) returns the dimension n of the block of A in my_id
!     npos(my_x) returns the columnindex of A which corresponds to 
!     first col. in the block in proc. with x pos. my_x 
  
npos(0)=1
temp=ceiling((real(n))/real(dim_x));
print *,"temp=",temp
do i=0,dim_x-2
   ndim(i)=temp
   npos(i+1)=npos(i)+temp
end do
ndim(dim_x-1)=n-npos(dim_x-1) + 1
npos(dim_x)=n+1

mpos(0)=1
temp=ceiling(real(m)/real(dim_y))
print *,"temp2=",temp
do i=0,dim_y-2
   mdim(i)=temp;
   mpos(i+1)=mpos(i)+temp;
end do
mdim(dim_y-1)=m-mpos(dim_y-1) + 1
mpos(dim_y)=m+1
  
my_m=mdim(my_y)
my_n=ndim(my_x)
  
if (my_id==0) then

   ! I cheat a little when distrubuting the matrices */
   
   allocate(A(1:my_m,1:my_n))
   allocate(B(1:my_m,1:my_n))
   allocate(C(1:my_m,1:my_n))
   
   A=1.5
   B=1.7
   C=0.0
   
   do i=1,numprocs-1
      
      ! If you want another A and B matris to be distributed
      ! you can define the parts here. 
      
      call MPI_CART_COORDS(cart_comm,i,CART_DIM,mycoord,ierr)
      my_x=mycoord(X)
      my_y=mycoord(Y)
!      print *,"sending to ",i
      call MPI_Send(A(1,1),mdim(my_y)*ndim(my_x),MPI_REAL,i,10,cart_comm,ierr)
      call MPI_Send(B(1,1),mdim(my_y)*ndim(my_x),MPI_REAL,i,20,cart_comm,ierr)
      
   end do
   call MPI_CART_COORDS(cart_comm,my_id,CART_DIM,mycoord,ierr)
   my_x=mycoord(X)
   my_y=mycoord(Y)
   
else 
   ! Allocate memory for the local matrices 
   ! This is a special structure bufferX(0) points on the
   !  data of the matrix X 
   
   allocate(A(1:my_m,1:my_n))
   allocate(B(1:my_m,1:my_n))
   allocate(C(1:my_m,1:my_n))
   C=0.0
   
   call MPI_RECV(A(1,1),my_m*my_n,MPI_REAL,0,10,cart_comm,status,ierr)
   call MPI_RECV(B(1,1),my_m*my_n,MPI_REAL,0,20,cart_comm,status,ierr)
end if
  
call MPI_BARRIER(cart_comm,ierr)

if (my_id==0) then
   t1=MPI_WTIME()
end if


!  -------------------------------------------------------------------
!  ---------------------  Make your routine here ---------------------
!  -------------------------------------------------------------------

!  The processors are arranged in a 2D grid and the links are specified
!  in the vector link, se below. The matrices is partitioned by the same
!  structure as the processors and distributed to corresponding proc.
!  A,B,C is the local block on each processor.

!  If you want a torus instead of a grid change NO_WRAP=0 to NO_WRAP=1
!  on line 17.

!     DIMM        : Number of column in A
!     my_m        : Number of rows in my partition of A 
!     my_n        : Number of columns in my partition of A
!                   my_m and my_n depends on the partition of processors
!                   they are usually not the same
!     my_x,my_y   : My x,y position in the proc. grid
!     link        : links to the neighborhood, use LEFT,RIGHT... as index
!                   if link do not exist than link()=MPI_PROC_NULL 
!                   (for example: if my_x=0 then link(Left)=MPI_PROC_NULL)
!     A,B,C       : My partition of A,B,C.
!                   A(my_mxmy_n), B(my_nxmy_m),C(my_mxmy_m). Index starts at 0
!     cart_comm   : communicator of the grid topology
!     mpos        : array 1:dim_x+1.  mpos(my_x) is the row index in the whole A 
!                   which corresponds to the first row in my partition A
!     npos        : array 1:dim_y+1 the same as mpos but index for columns insted.

!     link(LEFT)  : Backward in x 
!     link(RIGHT) : Forward  in x
!     link(UP)    : Backward in y
!     link(DOWN)  : Forward  in y


!  A typical MPI_Send call looks as follows 
!       call MPI_SEND(workA,size,MPI_REAL,link(xxx),tag,cart_com,ierr)
!  change the size, xxx and tag to correct values.

!  ---------- Write your matrix multiplication routine here ----------
!  ---------- The answer should be returned in the matrix C ----------
!  ----------  Use the algorithm 4.2 described on p. 109 in ----------
!  -------- Scientific Computing on High Performance Computers -------
!  ----------       by L. Elden, H. Park and Y. Saad        ----------

  
!  use these arrays as work arrays
allocate(workA(1:my_m))
allocate(workB(1:my_n))







! -------------------------------------------------------------------

if (my_id==0) then
   t1=MPI_WTIME()-t1;
   print *,"Total time:",t1
   print *,"Megaflops:",(2.0*n*n*n)/(t1*1.0e6)
end if
  
deallocate(A)
deallocate(B)
deallocate(C)
call MPI_FINALIZE(ierr);
end program SUMMA

