/*     Creates a 2d grid and do a matrix - matrix multiplication 
       between to matrices of the same dimensions using the MPI interface
*/
#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include "nrutil.h"

#define m 1000
#define n 1000
#define DIM n
#define true 1
#define false 0

#define REORDER  true  /* Reorder nodes if nessesary */
#define CART_DIM 2     /* Dimension of cartesian map */
#define NO_WRAP  0     /* No wrap around corners of the map */

#define UPDOWN   0
#define SIDEWAYS 1
#define LEFT	 2
#define RIGHT    3
#define DOWN	 1
#define UP       0
#define X        1
#define Y        0

double sekund()
{
  return MPI_Wtime();
}


void main(argc,argv)
int argc;
char *argv[];
{
  double t1;
  int my_id, numprocs, i,j,k;
  int  my_x, my_y, dim_x, dim_y,temp,my_m,my_n;
  int mdim[100],ndim[100],mpos[100],npos[100];
  int cartsize[2],period[2],mycoord[2],link[4];
  double **A, **B, **C;
  double *workA;
  double *workB;


  MPI_Comm cart_comm;
  MPI_Status status;

  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&my_id);
  

  /* Build a cartesian grid with optimal size */
  cartsize[X]=cartsize[Y]=0;
  MPI_Dims_create(numprocs,2,cartsize);
  dim_x=cartsize[X];
  dim_y=cartsize[Y];
  
  if (my_id==0){
    fprintf(stderr,"Dimension:%d\n",DIM);
    fprintf(stderr,"Number of proc:%d\n",numprocs);
    fprintf(stderr,"Dim x:%d  Dim y:%d\n",dim_x,dim_y);
  }

  
  period[X]=period[Y]=NO_WRAP;
  MPI_Cart_create(MPI_COMM_WORLD,CART_DIM,cartsize,
		  period,REORDER,&cart_comm);
  
  
  MPI_Comm_rank(MPI_COMM_WORLD,&my_id);
  MPI_Cart_coords(cart_comm,my_id,CART_DIM,mycoord);
  MPI_Cart_shift(cart_comm,SIDEWAYS,1,&link[LEFT],&link[RIGHT]);
  MPI_Cart_shift(cart_comm,UPDOWN,1,&link[UP],&link[DOWN]);

  my_x=mycoord[X];
  my_y=mycoord[Y];

  
  /*Distribute matrices */
  
  /*    Compute the dimensions of the partitions of A over the proc.
	Distrubute the matrix in blocks
	ndim(my_x) returns the dimension n of the block of A in my_id
	npos(my_x) returns the columnindex of A which corresponds to 
	first col. in the block in proc. with x pos. my_x */
  
  npos[0]=0;
  temp=ceil( ((float)n) /dim_x);
  for (i=0;i<=dim_x-2;i++){
    ndim[i]=temp;	
    npos[i+1]=npos[i]+temp;
  }
  ndim[dim_x-1]=n-npos[dim_x-1];
  npos[dim_x]=n+1;

  mpos[0]=0;
  temp=ceil( ((float)m) /dim_y);
  for(i=0;i<=dim_y-2;i++){
    mdim[i]=temp;
    mpos[i+1]=mpos[i]+temp;
  }
  mdim[dim_y-1]=m-mpos[dim_y-1];
  mpos[dim_y]=m+1;
  

  my_m=mdim[my_y];
  my_n=ndim[my_x];
  
  if (my_id==0){

    /* I will cheat a little when distrubuting the matrices */

    A=dmatrix(0,my_m-1,0,my_n-1);
    B=dmatrix(0,my_m-1,0,my_n-1);
    C=dmatrix(0,my_m-1,0,my_n-1);

    for (i=0;i<my_m;i++)
      for (j=0;j<my_n;j++){
	A[i][j]=1.0;
	B[i][j]=1.0;
	C[i][j]=0.0;
      }
    
    for (i=1;i<numprocs;i++){

      /* If you want another A and B matris to be distributed
	 you can define it here. */

      MPI_Cart_coords(cart_comm,i,CART_DIM,mycoord);
      my_x=mycoord[X];
      my_y=mycoord[Y];

      MPI_Send(A[0],mdim[my_y]*ndim[my_x],MPI_DOUBLE,i,10,cart_comm);
      MPI_Send(B[0],mdim[my_y]*ndim[my_x],MPI_DOUBLE,i,20,cart_comm);

    }
    MPI_Cart_coords(cart_comm,my_id,CART_DIM,mycoord);
    my_x=mycoord[X];
    my_y=mycoord[Y];
  }
  else {
    /* Allocate memory for the local matrices */
    /* This is a special structure bufferX[0] points on the
       data of the matrix X */

    A=dmatrix(0,my_m-1,0,my_n-1);
    B=dmatrix(0,my_m-1,0,my_n-1);
    C=dmatrix(0,my_m-1,0,my_n-1);
    for (i=0;i<my_m;i++)
      for (j=0;j<my_n;j++)
	C[i][j]=0.0;

    MPI_Recv (A[0],my_m*my_n,MPI_DOUBLE,0,10,cart_comm,&status);
    MPI_Recv (B[0],my_m*my_n,MPI_DOUBLE,0,20,cart_comm,&status);
  }
  
  MPI_Barrier(cart_comm);

  if (my_id==0)
    t1=sekund();
  
			
/*-------------------------------------------------------------------
  ---------------------  Make your routine here ---------------------
  -------------------------------------------------------------------

  The processors are arranged in a 2D grid and the links are specified
  in the vector link, se below. The matrices is partitioned by the same
  structure as the processors and distributed to corresponding proc.
  A,B,C is the local block on each processor.

  If you want a torus instead of a grid change NO_WRAP=0 to NO_WRAP=1
  on line 17.

     DIM         : Number of column in A
     my_m        : Number of rows in my partition of A 
     my_n        : Number of columns in my partition of A
                   my_m and my_n depends on the partition of processors
                   they are usually not the same
     my_x,my_y   : My x,y position in the proc. grid
     link        : links to the neighborhood, use LEFT,RIGHT... as index
                   if link do not exist than link()=MPI_PROC_NULL 
                   (for example: if my_x=0 then link(LEFT)=MPI_PROC_NULL)
     A,B,C       : My partition of A,B,C.
                   A(my_mxmy_n), B(my_nxmy_m),C(my_mxmy_m). Index starts at 0
     cart_comm   : communicator of the grid topology
     mpos        : array 0:dim_x.  mpos(my_x) is the row index in the whole A 
                   which corresponds to the first row in my partition A
     npos        : array 0:dim_y the same as mpos but index for columns insted.

     link(LEFT)  : Backward in x 
     link(RIGHT) : Forward  in x
     link(UP)    : Backward in y
     link(DOWN)  : Forward  in y


  A typical MPI_Send call looks as follows 
         MPI_Send(workA,size,MPI_DOUBEL,link[xxx],tag,cart_com)
  change the size, xxx and tag to correct values.

  ---------- Write your matrix multiplication routine here ----------
  ---------- The answer should be returned in the matrix C ----------
  ----------  Use the algorithm 4.2 described on p. 109 in  ---------
  -------- Scientific Computing on High Performance Computers -------
  ----------       by L. Elden, H. Park and Y. Saad        ----------*/

  
/* use these arrays as work arrays (usage workA[k]) */
  workA=dvector(0,my_m-1); 
  workB=dvector(0,my_n-1);





/*-------------------------------------------------------------------*/

if (my_id==0){
    t1=sekund()-t1;
    fprintf(stderr,"Total time:%g\n",t1);
    fprintf(stderr,"Megaflops:%g\n",(2.0*n*n*n)/(t1*1e6));
}

  free_dmatrix(A,0,my_m-1,0,my_n-1);
  free_dmatrix(B,0,my_m-1,0,my_n-1);
  free_dmatrix(C,0,my_m-1,0,my_n-1);
  MPI_Finalize();
}
